package main

import (
	"bytes"
	"crypto/rand"
	"encoding/json"
	"fmt"
	"log"
	"maps"
	"os"
	"os/exec"
	"path/filepath"
	"slices"
	"strings"
	"time"

	"github.com/alecthomas/kong"
	"golang.org/x/crypto/nacl/secretbox"
	"golang.org/x/crypto/scrypt"
	"golang.org/x/term"

	"git.jrop.me/jonathan/envvault/internal"
	"git.jrop.me/jonathan/envvault/internal/config"
	"git.jrop.me/jonathan/envvault/internal/daemon"
)

var cachedPassword []byte

const (
	EnvStoreCurrentVersion = 1
)

var (
	keyFilePath     string
	dbFilePath      string
	localDbFilePath string
)

func init() {
	homeDir, err := os.UserHomeDir()
	if err != nil {
		log.Fatalf("Failed to get user home directory: %v", err)
	}
	keyFilePath = filepath.Join(homeDir, ".local/cache/envvault/key")
	dbFilePath = filepath.Join(homeDir, ".local/cache/envvault/db.json.enc")
	localDbFilePath = ".envvault.json.enc"
}

type EnvStore struct {
	Vars map[string]string   `json:"vars"`
	Tags map[string][]string `json:"tags,omitempty"`
}

// GetTags returns all tags for a given variable name
func (s *EnvStore) GetTags(varName string) []string {
	if s.Tags == nil {
		return []string{}
	}
	return s.Tags[varName]
}

// SetTags sets the tags for a given variable name
func (s *EnvStore) SetTags(varName string, tags []string) {
	if s.Tags == nil {
		s.Tags = make(map[string][]string)
	}

	// Create a union of existing and new tags
	existingTags := s.Tags[varName]
	unionTags := make([]string, 0, len(existingTags)+len(tags))

	// Add existing tags to the union
	for _, tag := range existingTags {
		unionTags = append(unionTags, tag)
	}

	// Add new tags if they don't already exist
	for _, tag := range tags {
		if !slices.Contains(unionTags, tag) {
			unionTags = append(unionTags, tag)
		}
	}

	s.Tags[varName] = unionTags
}

// RemoveTags removes the specified tags from a variable
func (s *EnvStore) RemoveTags(varName string, tagsToRemove []string) {
	if s.Tags == nil || len(s.Tags[varName]) == 0 {
		return
	}

	existingTags := s.Tags[varName]
	remainingTags := make([]string, 0, len(existingTags))

	// Keep only tags that are not in tagsToRemove
	for _, tag := range existingTags {
		if !slices.Contains(tagsToRemove, tag) {
			remainingTags = append(remainingTags, tag)
		}
	}

	s.Tags[varName] = remainingTags
}

// HasAnyTag returns true if the variable has any of the specified tags
func (s *EnvStore) HasAnyTag(varName string, tags []string) bool {
	if len(tags) == 0 {
		return true // No tags specified, include all variables
	}

	varTags := s.GetTags(varName)
	for _, tag := range tags {
		if slices.Contains(varTags, tag) {
			return true
		}
	}
	return false
}

type CLI struct {
	Init   InitCmd   `cmd:"" help:"Initialize the vault."`
	List   ListCmd   `cmd:"" help:"List all environment variables."`
	Add    AddCmd    `cmd:"" help:"Add an environment variable."`
	Rm     RmCmd     `cmd:"" help:"Remove an environment variable."`
	Tag    TagCmd    `cmd:"" help:"Add or remove tags from an environment variable."`
	Exec   ExecCmd   `cmd:"" help:"Execute a command with environment variables."           alias:"x"`
	Rekey  RekeyCmd  `cmd:"" help:"Change the master password."`
	Daemon DaemonCmd `cmd:"" help:"Start the password caching daemon."`
	Lock   LockCmd   `cmd:"" help:"Lock the vault by stopping the password caching daemon."`
}

type InitCmd struct{}
type ListCmd struct {
	Values bool     `short:"v" help:"Show values of environment variables."             default:"false"`
	Local  bool     `short:"l" help:"List only from local vault."                       default:"false"`
	Tags   []string `short:"t" help:"Filter by tags (can be specified multiple times)."`
}
type AddCmd struct {
	Name  string `arg:"" help:"Name of the environment variable."`
	Value string `arg:"" help:"Value of the environment variable."    optional:""`
	Local bool   `       help:"Add to local vault instead of global."             short:"l" default:"false"`
}
type RmCmd struct {
	Name  string `arg:"" help:"Name of the environment variable to remove."`
	Local bool   `       help:"Remove from local vault only."               short:"l" default:"false"`
}
type TagCmd struct {
	Name   string   `arg:"" help:"Name of the environment variable."`
	Tags   []string `arg:"" help:"Tags to add or remove."              optional:""`
	Remove bool     `       help:"Remove tags instead of adding them."             short:"r" default:"false"`
	Local  bool     `       help:"Operate on local vault only."                    short:"l" default:"false"`
	List   bool     `       help:"List tags for the variable."                     short:"L" default:"false"`
}
type ExecCmd struct {
	Env   []string `short:"e" help:"Environment variables to set."`
	Local bool     `short:"l" help:"Use only local vault."                             default:"false"`
	Tags  []string `short:"t" help:"Filter by tags (can be specified multiple times)."`
	Cmd   string   `          help:"Command to execute."                                               arg:"" passthrough:"all"`
	Args  []string `          help:"Arguments for the command."                                        arg:""                   optional:""`
}
type RekeyCmd struct{}
type DaemonCmd struct {
	Timeout int `help:"Password cache timeout in minutes." default:"5"`
}
type LockCmd struct{}

func main() {
	internal.Log.Println("Starting envvault")

	cli := CLI{}
	ctx := kong.Parse(&cli)

	command := ctx.Command()
	internal.Log.Printf("Executing command: %s", command)

	switch command {
	case "init":
		subcommandInitVault()
	case "add <name>":
		subcommandAddEnvVar(cli.Add)
	case "add <name> <value>":
		subcommandAddEnvVar(cli.Add)
	case "rm <name>":
		subcommandRmEnvVar(cli.Rm.Name, cli.Rm.Local)
	case "list":
		subcommandListEnvVars(cli.List)
	case "exec <cmd>":
		subcommandExecCommand(cli.Exec)
	case "exec <cmd> <args>":
		subcommandExecCommand(cli.Exec)
	case "rekey":
		subcommandRekeyVault()
	case "daemon":
		subcommandStartDaemon(cli.Daemon.Timeout)
	case "lock":
		subcommandLockVault()
	case "tag <name>":
		subcommandTagEnvVar(cli.Tag)
	case "tag <name> <tags>":
		subcommandTagEnvVar(cli.Tag)
	default:
		internal.Log.Printf("Unknown command: %s", command)
		log.Fatal("Unknown command")
	}

	internal.Log.Printf("Command completed: %s", command)
}

func subcommandInitVault() { // {{{
	internal.Log.Println("Initializing vault")

	if _, err := os.Stat(keyFilePath); os.IsNotExist(err) {
		internal.Log.Println("Key file does not exist, creating new vault")

		fmt.Fprint(os.Stderr, "Enter a new master password: ")
		password, err := term.ReadPassword(int(os.Stdin.Fd()))
		fmt.Fprintln(os.Stderr)
		if err != nil {
			internal.Log.Printf("Failed to read password: %v", err)
			log.Fatal(err)
		}

		fmt.Fprint(os.Stderr, "Enter a new master password (again): ")
		password2, err := term.ReadPassword(int(os.Stdin.Fd()))
		fmt.Fprintln(os.Stderr)
		if err != nil {
			internal.Log.Printf("Failed to read confirmation password: %v", err)
			log.Fatal(err)
		}

		if !bytes.Equal(password, password2) {
			internal.Log.Println("Passwords do not match")
			log.Fatal("Passwords do not match")
		}
		internal.Log.Println("Passwords match, caching password")
		cachedPassword = password

		// Generate a random key
		internal.Log.Println("Generating random key")
		var key [32]byte
		if _, err := rand.Read(key[:]); err != nil {
			internal.Log.Printf("Failed to generate random key: %v", err)
			log.Fatal(err)
		}

		// Encrypt the key with the password
		internal.Log.Println("Encrypting key with password")
		encryptedKey := encryptKeyWithPassword(key[:], password)

		internal.Log.Printf("Creating key directory: %s", filepath.Dir(keyFilePath))
		if err := os.MkdirAll(filepath.Dir(keyFilePath), 0700); err != nil {
			internal.Log.Printf("Failed to create key directory: %v", err)
			log.Fatal(err)
		}

		internal.Log.Printf("Writing encrypted key to: %s", keyFilePath)
		if err := os.WriteFile(keyFilePath, encryptedKey, 0600); err != nil {
			internal.Log.Printf("Failed to write key file: %v", err)
			log.Fatal(err)
		}

		// Initialize the db file
		internal.Log.Println("Initializing empty environment store")
		emptyStore := &EnvStore{Vars: make(map[string]string)}
		saveGlobalEnvStore(emptyStore)

		internal.Log.Println("Vault initialized successfully")
		fmt.Fprintln(os.Stderr, "Vault initialized successfully.")
	} else {
		internal.Log.Println("Vault already initialized")
		fmt.Fprintln(os.Stderr, "Vault already initialized.")
	}
}

// }}}

func subcommandAddEnvVar(cmd AddCmd) { // {{{
	var store *EnvStore
	if cmd.Local {
		store = loadLocalEnvStore()
		internal.Log.Printf("Adding environment variable '%s' to local vault", cmd.Name)
	} else {
		store = loadGlobalEnvStore()
		internal.Log.Printf("Adding environment variable '%s' to global vault", cmd.Name)
	}

	if cmd.Value == "" {
		fmt.Fprintf(os.Stderr, "Enter value for %s: ", cmd.Name)
		inputValue, err := term.ReadPassword(int(os.Stdin.Fd()))
		fmt.Fprintln(os.Stderr)
		if err != nil {
			log.Fatal(err)
		}
		cmd.Value = string(inputValue)
	}
	store.Vars[cmd.Name] = cmd.Value

	if cmd.Local {
		saveLocalEnvStore(store)
		fmt.Fprintf(os.Stderr, "Environment variable '%s' added to local vault.\n", cmd.Name)
	} else {
		saveGlobalEnvStore(store)
		fmt.Fprintf(os.Stderr, "Environment variable '%s' added to global vault.\n", cmd.Name)
	}
} // }}}

func subcommandRmEnvVar(name string, localOnly bool) { // {{{
	localExists := false

	// Check local store if it exists
	if _, err := os.Stat(localDbFilePath); err == nil {
		localStore := loadLocalEnvStore()
		if _, exists := localStore.Vars[name]; exists {
			delete(localStore.Vars, name)
			saveLocalEnvStore(localStore)
			localExists = true
			fmt.Fprintf(os.Stderr, "Environment variable '%s' removed from local vault.\n", name)
		} else if localOnly {
			fmt.Fprintf(os.Stderr, "Environment variable '%s' not found in local vault.\n", name)
		}
	} else if localOnly {
		fmt.Fprintf(os.Stderr, "Local vault does not exist.\n")
	}

	// Check global store if not local-only
	if !localOnly {
		globalStore := loadGlobalEnvStore()
		if _, exists := globalStore.Vars[name]; exists {
			delete(globalStore.Vars, name)
			saveGlobalEnvStore(globalStore)
			fmt.Fprintf(os.Stderr, "Environment variable '%s' removed from global vault.\n", name)
		} else if !localExists {
			fmt.Fprintf(os.Stderr, "Environment variable '%s' not found in global vault.\n", name)
		}
	}
} // }}}

func subcommandTagEnvVar(cmd TagCmd) { // {{{
	var store *EnvStore

	// Determine which store to use
	if cmd.Local {
		if _, err := os.Stat(localDbFilePath); os.IsNotExist(err) {
			fmt.Fprintf(os.Stderr, "Local vault does not exist.\n")
			os.Exit(1)
		}
		store = loadLocalEnvStore()
		internal.Log.Printf("Using local vault")
	} else {
		store = loadGlobalEnvStore()
		internal.Log.Printf("Using global vault")
	}

	// Check if the variable exists
	if _, exists := store.Vars[cmd.Name]; !exists {
		vaultType := "global"
		if cmd.Local {
			vaultType = "local"
		}
		fmt.Fprintf(os.Stderr, "Environment variable '%s' not found in %s vault.\n",
			cmd.Name,
			vaultType)
		os.Exit(1)
	}

	// If list flag is set, just list the tags and return
	if cmd.List {
		tags := store.GetTags(cmd.Name)
		if len(tags) == 0 {
			fmt.Fprintf(os.Stderr, "No tags for '%s'\n", cmd.Name)
			return
		}
		for _, tag := range tags {
			fmt.Println(tag)
		}
		return
	}

	// If no tags provided and not in list mode, show error
	if len(cmd.Tags) == 0 {
		fmt.Fprintf(os.Stderr, "No tags provided. Use --list to view tags.\n")
		os.Exit(1)
	}

	// Process tags
	if cmd.Remove {
		internal.Log.Printf("Removing tags %v from variable '%s'", cmd.Tags, cmd.Name)
		store.RemoveTags(cmd.Name, cmd.Tags)
		fmt.Fprintf(
			os.Stderr,
			"Removed tags from '%s': %s\n",
			cmd.Name,
			strings.Join(cmd.Tags, ", "),
		)
	} else {
		internal.Log.Printf("Adding tags %v to variable '%s'", cmd.Tags, cmd.Name)
		store.SetTags(cmd.Name, cmd.Tags)
		fmt.Fprintf(os.Stderr, "Added tags to '%s': %s\n", cmd.Name, strings.Join(cmd.Tags, ", "))
	}

	// Save the store
	if cmd.Local {
		saveLocalEnvStore(store)
	} else {
		saveGlobalEnvStore(store)
	}
} // }}}

func subcommandListEnvVars(cmdArgs ListCmd) { // {{{
	var store *EnvStore

	// Determine which store to use
	if cmdArgs.Local {
		if _, err := os.Stat(localDbFilePath); os.IsNotExist(err) {
			fmt.Fprintf(os.Stderr, "Local vault does not exist.\n")
			return
		}
		store = loadLocalEnvStore()
	} else {
		// Default: load and merge stores
		store = loadEnvStore()
	}

	// Sort keys for consistent output
	var keys []string
	for k := range store.Vars {
		// Filter by tags if specified
		if store.HasAnyTag(k, cmdArgs.Tags) {
			keys = append(keys, k)
		}
	}
	slices.Sort(keys)

	for _, k := range keys {
		v := store.Vars[k]
		tags := store.GetTags(k)
		tagStr := ""
		if len(tags) > 0 {
			tagStr = fmt.Sprintf(" (tags: %s)", strings.Join(tags, ", "))
		}
		if cmdArgs.Values {
			fmt.Printf("%s=%s", k, v)
			fmt.Fprintf(os.Stderr, "%s", tagStr)
			fmt.Println()
		} else {
			fmt.Printf("%s", k)
			fmt.Fprintf(os.Stderr, "%s", tagStr)
			fmt.Println()
		}
	}
} // }}}

func subcommandExecCommand(cmdArgs ExecCmd) { // {{{
	var store *EnvStore

	// Determine which store to use
	if cmdArgs.Local {
		if _, err := os.Stat(localDbFilePath); os.IsNotExist(err) {
			fmt.Fprintf(os.Stderr, "Local vault does not exist.\n")
			os.Exit(1)
		}
		store = loadLocalEnvStore()
		internal.Log.Println("Using local vault only for execution")
	} else {
		// Default: load and merge stores
		store = loadEnvStore()
		internal.Log.Println("Using merged vaults for execution")
	}
	envVars := os.Environ()
	for candidateEnvName, candidateEnvValue := range store.Vars {
		// Skip if not matching tag filter
		if !store.HasAnyTag(candidateEnvName, cmdArgs.Tags) {
			continue
		}

		if len(cmdArgs.Env) == 0 {
			// If no env vars are specified, add all env vars
			envVars = append(envVars, fmt.Sprintf("%s=%s", candidateEnvName, candidateEnvValue))
			continue
		}
		if slices.Contains(cmdArgs.Env, candidateEnvName) {
			envVars = append(envVars, fmt.Sprintf("%s=%s", candidateEnvName, candidateEnvValue))
		}
	}

	cmd := exec.Command(cmdArgs.Cmd, cmdArgs.Args...)
	cmd.Env = envVars
	cmd.Stdout = os.Stdout
	cmd.Stderr = os.Stderr
	cmd.Stdin = os.Stdin

	if err := cmd.Run(); err != nil {
		log.Fatal(err)
	}
} // }}}

func subcommandStartDaemon(timeoutMinutes int) { // {{{
	internal.Log.Printf("Starting daemon with default timeout: %d minutes", timeoutMinutes)

	defaultTimeout := time.Duration(timeoutMinutes) * time.Minute
	timeout := config.GetDaemonTimeout(defaultTimeout)
	socketPath := daemon.GetSocketPath()
	internal.Log.Printf("Using socket path: %s", socketPath)

	// Ensure socket directory exists
	socketDir := filepath.Dir(socketPath)
	internal.Log.Printf("Creating socket directory: %s", socketDir)
	if err := os.MkdirAll(socketDir, 0700); err != nil {
		internal.Log.Printf("Failed to create socket directory: %v", err)
		log.Fatalf("Failed to create socket directory: %v", err)
	}

	fmt.Printf("Starting password daemon (timeout: %s)\n", timeout)
	fmt.Printf("Socket path: %s\n", socketPath)

	internal.Log.Println("Calling daemon.StartDaemon")
	// Start the daemon
	if err := daemon.StartDaemon(socketPath, timeout); err != nil {
		internal.Log.Printf("Failed to start daemon: %v", err)
		log.Fatalf("Failed to start daemon: %v", err)
	}
} // }}}

func subcommandLockVault() { // {{{
	internal.Log.Println("Locking vault by stopping daemon")

	socketPath := daemon.GetSocketPath()
	client := daemon.NewClient(socketPath)

	if !client.IsRunning() {
		internal.Log.Println("Daemon is not running, nothing to lock")
		fmt.Println("Vault is already locked (daemon not running)")
		return
	}

	internal.Log.Println("Sending kill command to daemon")
	err := client.KillDaemon()
	if err != nil {
		internal.Log.Printf("Failed to kill daemon: %v", err)
		log.Fatalf("Failed to lock vault: %v", err)
	}

	internal.Log.Println("Daemon stopped successfully")
	fmt.Println("Vault locked successfully")
} // }}}

func subcommandRekeyVault() { // {{{
	// Ask for current master password
	fmt.Fprint(os.Stderr, "Enter current master password: ")
	currentPassword, err := term.ReadPassword(int(os.Stdin.Fd()))
	fmt.Fprintln(os.Stderr)
	if err != nil {
		log.Fatal(err)
	}

	// Load and decrypt the master key with current password
	encryptedKey, err := os.ReadFile(keyFilePath)
	if err != nil {
		log.Fatal(err)
	}

	masterKey, err := decryptKeyWithPassword(encryptedKey, currentPassword)
	if err != nil {
		log.Fatal("Invalid password or corrupted key file")
	}

	// Ask for new master password
	fmt.Fprint(os.Stderr, "Enter new master password: ")
	newPassword, err := term.ReadPassword(int(os.Stdin.Fd()))
	fmt.Fprintln(os.Stderr)
	if err != nil {
		log.Fatal(err)
	}

	// Ask for new master password confirmation
	fmt.Fprint(os.Stderr, "Enter new master password (again): ")
	newPassword2, err := term.ReadPassword(int(os.Stdin.Fd()))
	fmt.Fprintln(os.Stderr)
	if err != nil {
		log.Fatal(err)
	}

	// Verify passwords match
	if !bytes.Equal(newPassword, newPassword2) {
		log.Fatal("Passwords do not match")
	}

	// Check if daemon is running and kill it
	socketPath := daemon.GetSocketPath()
	client := daemon.NewClient(socketPath)

	if client.IsRunning() {
		internal.Log.Println("Daemon is running, stopping it before rekey")
		fmt.Println("Stopping password daemon before changing master password...")
		err := client.KillDaemon()
		if err != nil {
			internal.Log.Printf("Failed to kill daemon: %v", err)
			log.Fatalf("Failed to stop daemon: %v", err)
		}
		internal.Log.Println("Daemon stopped successfully")
	}

	// Encrypt the master key with the new password
	newEncryptedKey := encryptKeyWithPassword(masterKey, newPassword)

	// Backup the current key file
	backupPath := keyFilePath + ".bak"
	if err := os.Rename(keyFilePath, backupPath); err != nil {
		log.Fatal(err)
	}
	fmt.Fprintf(os.Stderr, "Backed up key file to %s\n", backupPath)

	// Write the new encrypted key
	if err := os.WriteFile(keyFilePath, newEncryptedKey, 0600); err != nil {
		log.Fatal(err)
	}

	// Update cached password
	cachedPassword = newPassword

	fmt.Fprintln(os.Stderr, "Master password changed successfully.")
} // }}}

////////////////////////////////////////////////////////////////////////////////
//
// Utilities:
//
////////////////////////////////////////////////////////////////////////////////

func loadPassword() []byte { // {{{
	internal.Log.Println("Getting password without caching")

	// 1. Check for globally cached password
	if cachedPassword != nil {
		internal.Log.Println("Using cached password from memory")
		return cachedPassword
	}

	internal.Log.Println("No cached password in this process' memory, trying daemon")

	// 2. Try to get password from daemon
	socketPath := daemon.GetSocketPath()
	internal.Log.Printf("Using socket path: %s", socketPath)
	client := daemon.NewClient(socketPath)

	err := client.EnsureRunning()
	if err != nil {
		internal.Log.Printf("Failed to spawn daemon: %v", err)
		log.Fatal(err)
	}

	// Check if daemon is running
	internal.Log.Println("Attempting to retrieve password from daemon")
	password, err := client.RetrievePassword()
	if err == nil {
		internal.Log.Println("Successfully retrieved password from daemon")
		return password
	}
	internal.Log.Printf("Failed to retrieve password from daemon: %v", err)

	// 3. Fall back to terminal input
	internal.Log.Println("Falling back to terminal input for password")
	fmt.Fprint(os.Stderr, "Enter master password: ")
	password, err = term.ReadPassword(int(os.Stdin.Fd()))
	if err != nil {
		internal.Log.Printf("Failed to read password from terminal: %v", err)
		log.Fatal(err)
	}
	fmt.Fprintln(os.Stderr) // Ensure newline after password input

	return password
} // }}}

func loadKey() [32]byte { // {{{
	// Get password but don't cache it yet
	password := loadPassword()

	encryptedKey, err := os.ReadFile(keyFilePath)
	if err != nil {
		log.Fatal(err)
	}

	decryptedKey, err := decryptKeyWithPassword(encryptedKey, password)
	if err != nil {
		log.Fatal("Invalid password or corrupted key file")
	}

	var key [32]byte
	copy(key[:], decryptedKey)

	// Only cache the password after we've verified it works

	internal.Log.Println("Caching validated password")

	// Cache in memory
	cachedPassword = password

	// Try to start daemon if not running
	socketPath := daemon.GetSocketPath()
	client := daemon.NewClient(socketPath)

	err = client.EnsureRunning()
	if err != nil {
		internal.Log.Printf("Failed to spawn daemon: %v", err)
		// If the daemon doesn't start, we don't need to stop the current process
		return key
	}

	// Store in daemon if it's running
	internal.Log.Println("Storing password in daemon")
	if err := client.StorePassword(password); err != nil {
		internal.Log.Printf("Failed to store password in daemon: %v", err)
		log.Printf("Failed to store password in daemon: %v", err)
		// Even if we have an error, we still want to return the key
		return key
	}

	internal.Log.Println("Successfully stored password in daemon")

	return key
} // }}}

func encryptKeyWithPassword(key, password []byte) []byte { // {{{
	var nonce [24]byte
	if _, err := rand.Read(nonce[:]); err != nil {
		log.Fatal(err)
	}

	passwordDerivedKey, err := scrypt.Key(password, nonce[:], 32768, 8, 1, 32)
	if err != nil {
		log.Fatal(err)
	}

	encrypted := secretbox.Seal(nonce[:], key, &nonce, (*[32]byte)(passwordDerivedKey))
	return encrypted
} // }}}

func decryptKeyWithPassword(encryptedKey, password []byte) ([]byte, error) { // {{{
	var nonce [24]byte
	copy(nonce[:], encryptedKey[:24])
	encrypted := encryptedKey[24:]

	passwordDerivedKey, err := scrypt.Key(password, nonce[:], 32768, 8, 1, 32)
	if err != nil {
		return nil, err
	}

	decrypted, ok := secretbox.Open(nil, encrypted, &nonce, (*[32]byte)(passwordDerivedKey))
	if !ok {
		return nil, fmt.Errorf("decryption failed")
	}

	return decrypted, nil
} // }}}

func loadStoreFile(filePath string) *EnvStore { // {{{
	key := loadKey()

	data, err := os.ReadFile(filePath)
	if err != nil {
		if os.IsNotExist(err) {
			return &EnvStore{
				Vars: make(map[string]string),
				Tags: make(map[string][]string),
			}
		}
		log.Fatal(err)
	}

	var store EnvStore
	if len(data) > 0 {
		var nonce [24]byte
		copy(nonce[:], data[:24])
		decrypted, ok := secretbox.Open(nil, data[24:], &nonce, &key)
		if !ok {
			log.Fatal("Decryption failed")
		}
		if err := json.Unmarshal(decrypted, &store); err != nil {
			log.Fatal(err)
		}
		// Initialize Tags map if it's nil
		if store.Tags == nil {
			store.Tags = make(map[string][]string)
		}
	} else {
		return &EnvStore{
			Vars: make(map[string]string),
			Tags: make(map[string][]string),
		}
	}
	return &store
} // }}}

func loadGlobalEnvStore() *EnvStore { // {{{
	if err := os.MkdirAll(filepath.Dir(dbFilePath), 0700); err != nil {
		log.Fatal(err)
	}
	return loadStoreFile(dbFilePath)
} // }}}

func loadLocalEnvStore() *EnvStore { // {{{
	return loadStoreFile(localDbFilePath)
} // }}}

func loadEnvStore() *EnvStore { // {{{
	// Load global store first
	globalStore := loadGlobalEnvStore()

	// Check if local store exists
	if _, err := os.Stat(localDbFilePath); os.IsNotExist(err) {
		// No local store, just return global
		return globalStore
	}

	// Load local store
	localStore := loadLocalEnvStore()

	// Merge stores (local takes precedence)
	mergedStore := &EnvStore{
		Vars: make(map[string]string),
		Tags: make(map[string][]string),
	}

	// Copy all global vars and tags
	maps.Copy(mergedStore.Vars, globalStore.Vars)
	maps.Copy(mergedStore.Tags, globalStore.Tags)

	// Override/add local vars and tags
	maps.Copy(mergedStore.Vars, localStore.Vars)
	maps.Copy(mergedStore.Tags, localStore.Tags)

	return mergedStore
} // }}}

func saveGlobalEnvStore(store *EnvStore) { // {{{
	key := loadKey()
	data, err := json.Marshal(store)
	if err != nil {
		log.Fatal(err)
	}

	var nonce [24]byte
	if _, err := rand.Read(nonce[:]); err != nil {
		log.Fatal(err)
	}
	encrypted := secretbox.Seal(nonce[:], data, &nonce, &key)

	if err := os.WriteFile(dbFilePath, encrypted, 0600); err != nil {
		log.Fatal(err)
	}
} // }}}

func saveLocalEnvStore(store *EnvStore) { // {{{
	key := loadKey()
	data, err := json.Marshal(store)
	if err != nil {
		log.Fatal(err)
	}

	var nonce [24]byte
	if _, err := rand.Read(nonce[:]); err != nil {
		log.Fatal(err)
	}
	encrypted := secretbox.Seal(nonce[:], data, &nonce, &key)

	if err := os.WriteFile(localDbFilePath, encrypted, 0600); err != nil {
		log.Fatal(err)
	}
} // }}}
